#=======================================================================
#	$Id: Model.pm,v 1.8 2007/04/05 10:05:32 pythontech Exp $
#	Model based on triple store.
#	Copyright (C) 2004-2007  Python Technology Limited
#
#	This program is free software; you can redistribute it and/or
#	modify it under the terms of the GNU General Public License
#	as published by the Free Software Foundation; either version 2
#	of the License, or (at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
#	02111-1307, USA.
#-----------------------------------------------------------------------
#	In the world seen by users of the model, there are:
#	  PythonTech::Obj (or subclass)		resource (URI = $obj->uri)
#	  \$uri					resource
#	  scalar				literal
#=======================================================================
package PythonTech::Model;
use strict;
use PythonTech::Obj;

=head1 NAME

PythonTech::Model - knowledge model based on triplestore

=head1 SYNOPSIS

  use PythonTech::Model;

  $model = PythonTech::Model->new(
    triples => $triples,
    classifier => $classifier,
    classmap => {
      '' => $defaultclass,
      $type1 => $perlclass1,
      $type2 => $perlclass2,
      ...
    },
    extqmap => $xqmap,
  );

  @objs = $model->all_of_type($type);

  $obj = $model->as_obj($uri);

  $obj2 = $model->create_obj($uri, $type,
			     $attr1 => $value1,
			     $attr2 => $value2, ...);

  $model->add($subj, $pred, $obj);

  $model->remove($subj, $pred, $obj);

  $tf = $model->test($subj, $pred, $obj);

  @obj = $model->sp2o($subj, $pred);

  @subj = $model->po2s($subj, $pred, $obj);

  @res = $model->query($subj, $pred, $obj);

  $model->commit;

  $model->rollback;

  $model->disconnect;

  $model->reconnect;

=head1 DESCRIPTION

This module implements a RDF knowledge model using a triplestore as
the underlying storage.

=cut

#--- Well-known URIs for standard stuff
my $rdf = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#';
my $rdftype = $rdf.'type';

#-----------------------------------------------------------------------

=head1 CONSTRUCTOR

  $model->PythonTech::Model->new(name => value, ...)

Create new model.  Configuration options may be given either
as a hash or as many name=>value arguments

The configuration options are:

=over 4

=item triples

Underlying triple store e.g. an instance of PythonTech::Triples::DBHash

=item classifier

(Optional) object which maps URIs to Perl classes.  If defined, the model
calls this object's B<uri2class> method to obtain a Perl class into which to 
bless objects created for URIs in the triplestore.  The returnd class should
normally be a subclass of B<PythonTech::Obj>.

=item classmap

(Optional) hash which maps from RDF type to Perl class.
Used if B<classifier> is not defined.  An entry whose key is the empty
string (if present) indicates a default class to use.

=item extqmap

(Optional) object which maps between qnames used for URIs externally and
the true URI space.  Normally an instance of PythonTech::QMap.

=item dbqmap

(Optional) object which mas between qnames used within the triplestore
and the true URI space.  Normally an instance of PythonTech::QMap.  This
option is deprecated - use PythonTech::Triples::QMapped instead to wrap
the triplestore.

=back

=cut

#-----------------------------------------------------------------------
sub new {
    my $class = shift;
    my $self = {
	triples => {},		# PythonTech::Triples
	classmap => {},		# { $rdftype => $perlclass }
	objmap => {},		# { $uri => $obj }
	dbrdftype => $rdftype,	# qURI for rdf:type in triples
    };
    bless $self, $class;
    $self->configure(@_);
    return $self;
}

sub configure {
    my $self = shift;
    my @props = ref($_[0]) eq 'HASH' ? %{$_[0]} : @_;
    while (my($key,$value) = splice(@props,0,2)) {
	if ($key eq 'triples') {
	    $self->{'triples'} = $value;
	} elsif ($key eq 'dbqmap') {
	    $self->{'dbqmap'} = $value;
	    if (defined $value) {
		$self->{'dbrdftype'} = $value->encode($rdftype);
	    }
	} elsif ($key eq 'extqmap') {
	    $self->{'extqmap'} = $value;
	} elsif ($key eq 'classifier') {
	    $self->{'classifier'} = $value;
	} elsif ($key eq 'classmap') {
	    # Augment any existing class mappings
	    keys %$value;
	    while (my($class,$perlclass) = each %$value) {
		$self->{'classmap'}->{$class} = $perlclass;
	    }
	} else {
	    die ref($self).": Unknown configuration option $key\n";
	}
    }
}

=head1 METHOD ARGUMENTS

For query methods, the supplied arguments should be of the following forms:

=over 4

=item $subj

Can be a PythonTech::Obj (or subclass), or a reference to a scalar URI,
or a scalar URI.

=item $pred

Must be a scalar URI.

=item $obj

For a resource, can be a PythonTech::Obj (or subclass), or a reference
to a scalar URI; for a literal, must be a scalar.

=back

=head1 METHODS

=over 4

=cut

#-----------------------------------------------------------------------

=item I<all_of_type($type)>

Return all resources with given RDF type

=cut

#-----------------------------------------------------------------------
sub all_of_type {
    my($self, $type) = @_;
    return $self->po2s($rdftype,\$type);
}

#-----------------------------------------------------------------------

=item I<as_obj($uri)>

Given a resource I<$uri>, instantiate it as an object.  The Perl class
into which the object is blessed is determined thus:
If a I<classifier> is configured, call its B<uri2class> method to get
the Perl class; else if a I<classmap> is defined, look up each rdf:type
defined for I<$uri> in the model in the map; else bless into
B<PythonTech::Obj>.

=cut

#-----------------------------------------------------------------------
sub as_obj {
    my($self, $xuri) = @_;
#    print STDERR "as_obj: $uri\n";
    my $uri = $xuri;
    if ($self->{'extqmap'}) {
	$uri = $self->{'extqmap'}->decode($xuri);
	$xuri = $self->{'extqmap'}->encode($uri);
    }
    # See if already known
    my $obj = $self->{'objmap'}->{$uri};
    return $obj if defined($obj);

    # Create new object, as yet unblessed
    $obj = {
	uri => $xuri,
	model => $self,
    };

    # Bless it into appropriate class, to get extra semantics
    if ($self->{'classifier'}) {
	# External object classifies according to URI
	my $perlclass = $self->{'classifier'}->uri2class($uri)
	    || 'PythonTech::Obj';
	eval "use $perlclass; 1"
	    or die "Cannot find perl class $perlclass: $@\n";
	bless $obj, $perlclass;
    } else {
	# Check if it is defined (i.e. has an rdf:type)
	my $turi = $uri;
	if ($self->{'dbqmap'}) {
	    $turi = $self->{'dbqmap'}->encode($turi);
	}

	my @types = $self->{'triples'}->sp2o($turi, $self->{'dbrdftype'});

	# Create it in base class
	bless $obj, 'PythonTech::Obj';

	# Bless it into the appropriate Perl class.
	# There may be more than one (subclassing) - make sure it is in the
	# most specific. (Beware - random choice if multiple inheritance!)
	foreach my $type ('',@types) {
	    my $xtype = $type;
	    if ($self->{'dbqmap'}) {
		$xtype = $self->{'dbqmap'}->decode($xtype);
	    }
	    if ($self->{'extqmap'}) {
		$xtype = $self->{'extqmap'}->encode($xtype);
	    }
#	    print STDERR " type=$type xtype=$xtype\n";
	    if (defined(my $perlclass = $self->{'classmap'}->{$xtype})) {
#		print STDERR "as_obj - perlclass $perlclass\n";
#		print STDERR " perlclass=$perlclass\n";
		unless ($obj->isa($perlclass)) {
		    eval "use $perlclass; 1"
			or die "Cannot find perl class $perlclass: $@\n";
		    bless $obj, $perlclass;
#		    print STDERR "as_obj - blessed as $perlclass\n";
		}
	    }
	}
    }

    # Cache it
    $self->{'objmap'}->{$uri} = $obj;
#    print STDERR "After class blessing\n";
    return $obj;
}

#-----------------------------------------------------------------------

=item create_obj($uri, $type, $pred1 => $obj1, ...)

Create a new object

=cut

#-----------------------------------------------------------------------
sub create_obj {
    my($self, $uri,$type,@attr) = @_;
    $self->add(\$uri,$rdftype,\$type);
    while (my($pred,$obj) = splice(@attr,0,2)) {
	$self->add(\$uri,$pred,$obj);
    }
    return $self->as_obj($uri);
}

#-----------------------------------------------------------------------

=item I<add($subj,$pred,$obj)>

Add a statement to the model

=cut

#-----------------------------------------------------------------------
sub add {
    my($self, $subj,$pred,$obj) = @_;
    my $tsubj = $self->_insub($subj);
    my $tpred = $self->_inpred($pred);
    my $tobj = $self->_in($obj);
    $self->{'triples'}->add($tsubj,$tpred,$tobj);
}

#-----------------------------------------------------------------------

=item I<remove($subj,$pred,$obj)>

Remove a statement from the model

=cut

#-----------------------------------------------------------------------
sub remove {
    my($self, $subj,$pred,$obj) = @_;
    my $tsubj = $self->_insub($subj);
    my $tpred = $self->_inpred($pred);
    my $tobj = $self->_in($obj);
    $self->{'triples'}->remove($tsubj,$tpred,$tobj);
}

#-----------------------------------------------------------------------

=item I<test($subj,$pred,$obj)>

Returns 1 if the model contains the statement, else undef.

=cut

#-----------------------------------------------------------------------
sub test {
    my($self, $subj,$pred,$obj) = @_;
    my $tsubj = $self->_insub($subj);
    my $tpred = $self->_inpred($pred);
    my $tobj = $self->_in($obj);
    return $self->{'triples'}->test($tsubj,$tpred,$tobj);
}

#-----------------------------------------------------------------------

=item I<sp2o($subj,$pred)>

Get object(s) given subject and predicate.  In array context, returns
all matching objects; in scalar context, returns the first (if any).

=cut

#-----------------------------------------------------------------------
sub sp2o {
    my($self, $subj,$pred) = @_;
    my $tsubj = $self->_insub($subj);
    my $tpred = $self->_inpred($pred);
#    print STDERR "sp2o: subj=$subj tsubj=$tsubj pred=$pred tpred=$tpred\n";
    if (wantarray) {
	my @tobjs = $self->{'triples'}->sp2o($tsubj,$tpred);
	return map {$self->_out($_)} @tobjs;
    } else {
	my $tobj = $self->{'triples'}->sp2o($tsubj,$tpred);
	return defined($tobj) ? $self->_out($tobj) : undef;
    }
}

#-----------------------------------------------------------------------

=item I<po2s($pred,$obj)>

Get subject(s) given predicate and object.  In array context, returns
all matching subjects; in scalar context, returns the first (if any).

=cut

#-----------------------------------------------------------------------
sub po2s {
    my($self, $pred,$obj) = @_;
    my $tpred = $self->_inpred($pred);
    my $tobj = $self->_in($obj);

    if (wantarray) {
	my @tsubjs = $self->{'triples'}->po2s($tpred,$tobj);
	return map {$self->_out($_)} @tsubjs;
    } else {
	my $tsubj = $self->{'triples'}->po2s($tpred,$tobj);
	return defined($tsubj) ? $self->_out($tsubj) : undef;
    }
}

#-----------------------------------------------------------------------

=item I<query($subj,$pred,$obj)>

Perform a generic query for statements matching those of $subj, $pred
and $obj which are defined.  Each returned value is an array ref 
containing the missing items.  E.g. if $subj is defined and $pred and
$obj are undef, each result is an array ref containing a predicate and
an object.  In array context returns all matching results; in scalar
context returns the first result (if any).

=cut

#-----------------------------------------------------------------------
sub query {
    my($self, $subj,$pred,$obj) = @_;
    my $tsubj = defined $subj ? $self->_insub($subj) : undef;
    my $tpred = defined $pred ? $self->_inpred($pred) : undef;
    my $tobj  = defined $obj  ? $self->_in($obj)  : undef;
    if (wantarray) {
	my @res = $self->{'triples'}->query($tsubj,$tpred,$tobj);
	return map {[map {$self->_out($_)} @$_]} @res;
    } else {
	my $res = $self->{'triples'}->query($tsubj,$tpred,$tobj);
	return defined($res) ? [map {$self->_out($_)} @$res] : undef;
    }
}

#-----------------------------------------------------------------------

=item I<commit>

Commit any changes which are not yet made permanent in the underlying
triplestore.

=cut

#-----------------------------------------------------------------------
sub commit {
    my($self) = @_;
    $self->{'triples'}->commit;
}

#-----------------------------------------------------------------------

=item I<rollback>

Discard any changes which are not yet made permanent in the underlying
triplestore.

=cut

#-----------------------------------------------------------------------
sub rollback {
    my($self) = @_;
    $self->{'triples'}->rollback;
}

#-----------------------------------------------------------------------

=item I<refresh([$force])>

Ask the triplestore to check if its data has become out-of-date
with respect to its source data.  If I<$force> is true, refresh even if
it is expensive to do.

=cut

#-----------------------------------------------------------------------
sub refresh {
    my($self,@args) = @_;
    if ($self->{'triples'}->can('refresh')) {
	$self->{'triples'}->refresh(@args);
    }
}

#-----------------------------------------------------------------------
#	Disconnect / reconnect from underlying triplestore
#-----------------------------------------------------------------------
sub disconnect {
    my $self = shift;
    $self->{'triples'}->disconnect;
}

sub reconnect {
    my $self = shift;
    $self->{'triples'}->reconnect;
}

#-----------------------------------------------------------------------
#	Convert subject to URI
#-----------------------------------------------------------------------
sub _insub {
    my($self, $value) = @_;
    my $uri;
    if (ref($value) eq 'SCALAR') {
	$uri = $$value;
    } elsif (ref($value)) {
	$uri = eval {$value->uri};
	unless ($uri) {
	    die "no uri: $value ".ref($value)."";
	}
    } else {
	$uri = $value;
    }
    my $quri = $uri;
    if ($self->{'extqmap'}) {
	$quri = $self->{'extqmap'}->decode($quri);
    }
    if ($self->{'dbqmap'}) {
	$quri = $self->{'dbqmap'}->encode($quri);
    }
    return $quri;
}

#-----------------------------------------------------------------------
#	Convert predicate to database standard URI
#-----------------------------------------------------------------------
sub _inpred {
    my($self, $uri) = @_;
    my $turi = $uri;
    if ($self->{'extqmap'}) {
	$turi = $self->{'extqmap'}->decode($turi);
    }
    if ($self->{'dbqmap'}) {
	$turi = $self->{'dbqmap'}->encode($turi);
    }
    return $turi;
}

#-----------------------------------------------------------------------
#	Convert object or literal to URI or =literal
#-----------------------------------------------------------------------
sub _in {
    my($self, $value) = @_;
    my $uri;
    if (ref($value) eq 'SCALAR') {
	$uri = $$value;
    } elsif (ref($value)) {
	$uri = eval {$value->uri};
	unless ($uri) {
	    die "no uri: $value ".ref($value)."";
	}
    } else {
	return "=$value";
    }
    my $quri = $uri;
    if ($self->{'extqmap'}) {
	$quri = $self->{'extqmap'}->decode($quri);
    }
    if ($self->{'dbqmap'}) {
	$quri = $self->{'dbqmap'}->encode($quri);
    }
    return $quri;
}

#-----------------------------------------------------------------------
#	Convert value from triple store to object or literal
#-----------------------------------------------------------------------
sub _out {
    my($self, $value) = @_;
    if ($value =~ /^=/o) {
	return substr($value,1);
    } else {
	my $uri = $value;
	if ($self->{'dbqmap'}) {
	    $uri = $self->{'dbqmap'}->decode($uri);
	}
	if ($self->{'extqmap'}) {
	    $uri = $self->{'extqmap'}->encode($uri);
	}
	return $self->as_obj($uri);
    }
}

1;
